
local Whammy = CreateFrame("FRAME", nil, UIParent);

Whammy.PERFORMANCE_FACTOR = 0.045;  -- performance throttle (bigger is less frequent)
Whammy.VANISH_TIMER = 2.5;          -- fadeout speed (bigger is slower)


-- cache player class immediately (available during load)
_,Whammy.Class = UnitClass("player");

-- cache state information during Whammy:InventoryWatch()
Whammy.DualWield = nil;
Whammy.Ranged = nil;

-- relevant inventory slots
Whammy.MAINHAND_SLOT = GetInventorySlotInfo("MainHandSlot");
Whammy.OFFHAND_SLOT = GetInventorySlotInfo("SecondaryHandSlot");
-- ranged slot no longer exists

-- relevant spell ids
Whammy.RANGED_SPELL = {
  [75] = true,   -- Hunter AutoShot
  [5019] = true  -- Mage/Priest/Warlock Shoot Wand
}

-- relevant handedness strings
-- TODO: replace with non-string based detection if possible (should fix for all localizations)
local DETECT_DUALWIELD = {};
DETECT_DUALWIELD["One-Hand"] = true;
DETECT_DUALWIELD["Two-Hand"] = true;
DETECT_DUALWIELD["Off Hand"] = true;


-- KEEP: this snippet may come in handy for a future idea
--[[
local function CanPlayerCast(spellid)
  name = GetSpellInfo(spellid);
  if ( not name ) then
    return(nil);
  end
  return(1);
end
]]--



-- swing animation frame --
-- --------------------- --

local function SetSwingBarWidth(element, time, speed, maxwidth)
  local width = ( time / speed * maxwidth );
  if ( width < 1 ) then
    width = 1;
  elseif ( width > maxwidth ) then
    width = maxwidth;
  end
  element:SetWidth(width);
end


local function GetSwingSpeed(self)

  local mainhand,offhand = UnitAttackSpeed("player");

-- no longer seems to be necessary
  -- special case: huntards
--[[
  if ( ( Whammy.Ranged ) and ( Whammy.Class == "HUNTER" ) ) then
    return(mainhand-(mainhand*0.15));  -- why isn't this just factored in
  end
]]--

  if ( Whammy.DualWield ) and ( self.SLOT == Whammy.OFFHAND_SLOT ) then
    return(offhand);
  else
    return(mainhand);
  end
end


local function SetSwingSpeed(self, speed)
  if ( speed == nil ) then
    self.speed = GetSwingSpeed(self);
    if ( self.speed == nil ) then
      Whammy.Disable(self);
      return(nil);
    end
  else
    self.speed = speed;
  end

  self.text:SetText(string.format("%.2f", math.floor(self.speed * 100 + 0.5) / 100));

  _,_,self.latency = GetNetStats();
  self.latency = self.latency / 1000;
  SetSwingBarWidth(self.lagbar, self.latency, self.speed, self.maxwidth);
end


local function SwingFrameReset(self)
  self:SetScript("OnUpdate", nil);
  SetSwingSpeed(self);
  self.lastupdate = 0;
  self.elapsed = 0;
  self.bar:SetWidth(1);
  self.active = nil;
  self:Show();
  self.vanish = 0;
end


local function SwingFrameVanish(self, elapsed)
  if ( self.vanish > 0 ) then
    self.vanish = self.vanish - elapsed;
    if ( self.vanish <= 0 ) then
      self:SetScript("OnUpdate", nil);
      self:Hide();
    end
  end  
end


local function SwingFrameTimer(self, elapsed)
  -- performance throttle
  --
  self.lastupdate = self.lastupdate + elapsed;
  if ( self.lastupdate < Whammy.PERFORMANCE_FACTOR ) then
    return(nil);
  end
  self.elapsed = self.elapsed + self.lastupdate;
  self.lastupdate = 0;

  -- flag the swing timer as inactive early to help account for lag
  --
  if ( self.elapsed > (self.speed - self.latency ) ) then
    self.active = nil;
  end

  if ( self.elapsed > self.speed ) then
    SwingFrameReset(self);
    self.vanish = Whammy.VANISH_TIMER;
    self:SetScript("OnUpdate", SwingFrameVanish);
  end

  SetSwingBarWidth(self.bar, self.elapsed, self.speed, self.maxwidth);
end


local function CreateSwingFrame(id)
  local frame = CreateFrame("FRAME", nil, UIParent);
  frame:Hide();
  frame.maxwidth = 200;
  frame.id = id;
  
  frame:SetWidth(frame.maxwidth);
  frame:SetHeight(5);

  frame.backdrop = frame:CreateTexture(nil, "BACKGROUND");
  frame.backdrop:SetAllPoints();
  frame.backdrop:SetTexture(1,1,1,0.1);

  frame.bar = frame:CreateTexture(nil, "ARTWORK");
  frame.bar:SetPoint("LEFT", frame.backdrop);
  frame.bar:SetWidth(1);
  frame.bar:SetHeight(3);
  frame.bar:SetTexture(0.8,0.8,0.8,0.8);

  frame.lagbar = frame:CreateTexture(nil, "OVERLAY");
  frame.lagbar:SetPoint("RIGHT", frame.backdrop);
  frame.lagbar:SetWidth(1);
  frame.lagbar:SetHeight(3);
  frame.lagbar:SetTexture(0.7,0,0,0.4);

  frame.left = frame:CreateTexture(nil, "OVERLAY");
  frame.left:SetPoint("LEFT", frame.backdrop, -1);
  frame.left:SetWidth(2);
  frame.left:SetHeight(5);
  frame.left:SetTexture(0.2,0.2,0.2,1);

  frame.right = frame:CreateTexture(nil, "OVERLAY");
  frame.right:SetPoint("RIGHT", frame.backdrop, 1);
  frame.right:SetWidth(2);
  frame.right:SetHeight(5);
  frame.right:SetTexture(0.2,0.2,0.2,1);

  frame.text = frame:CreateFontString(nil, "ARTWORK");
  frame.text:SetPoint("RIGHT", frame.left, "LEFT", -4, 1);
  frame.text:SetFont("Fonts\\FRIZQT__.TTF", 8);
  frame.text:SetTextColor(1,1,1,1);

  return(frame);
end



-- Inventory Detection Frame --
-- ------------------------- --

CreateFrame( "GameTooltip", "WhammyTooltip", nil, "GameTooltipTemplate" );
WhammyTooltip:SetOwner( WorldFrame, "ANCHOR_NONE" );
WhammyTooltip:AddFontStrings(
  WhammyTooltip:CreateFontString( "$parentTextLeft1", nil, "GameTooltipText" ),
  WhammyTooltip:CreateFontString( "$parentTextLeft9", nil, "GameTooltipText" ),
  WhammyTooltip:CreateFontString( "$parentTextRight1", nil, "GameTooltipText" )
);


local function InventoryDetect(slot, detect)
  local text;
  WhammyTooltip:ClearLines();
  WhammyTooltip:SetInventoryItem("player", slot);
  for i=1,4 do
    text = _G["WhammyTooltipTextLeft"..i]:GetText();
    if ( text ~= nil ) and ( detect[text] ) then
      return(1);
    end
  end
  return(nil);
end


function PlayerWeaponIsDualWield()
  return(InventoryDetect(Whammy.OFFHAND_SLOT, DETECT_DUALWIELD));
end  


function PlayerWeaponIsRanged()
  ranged = UnitRangedDamage("player");
  if ( ranged > 0 ) then
    return(1);
  else
    return(nil);
  end
end


-- swing frames --
-- ------------ --

local SwingFrame = {};
SwingFrame[1] = CreateSwingFrame(1);
SwingFrame[2] = CreateSwingFrame(2);
--SwingFrame[3] = CreateSwingFrame(3);

SwingFrame[1]:SetPoint("TOP", CastingBarFrame, "BOTTOM", 0, -2);
SwingFrame[2]:SetPoint("TOP", SwingFrame[1], "BOTTOM", 0, -2);
--SwingFrame[3]:SetPoint("TOP", SwingFrame[2], "BOTTOM", 0, -2);



-- swing event handlers --
-- -------------------- --


local function SwingFrameMainHandEvent(self,_,_,event,_,src_guid)
  if ( string.sub(event,1,5) == "SWING" ) and ( src_guid == UnitGUID("player") ) then
    SwingFrameReset(self);
    self:SetScript("OnUpdate", SwingFrameTimer);
  end
end


--local function SwingFrameRangedEvent(self,_,_,event,_,src_guid)
local function SwingFrameRangedEvent(self,_,_,event,_,src_guid,_,_,_,_,_,_,_,spellID)
  if ( Whammy.RANGED_SPELL[spellID] ~= true ) then return; end

  if ( string.sub(event,1,5) == "RANGE" ) and ( src_guid == UnitGUID("player") ) then
    SwingFrameReset(self);
    self:SetScript("OnUpdate", SwingFrameTimer);
  end
end


-- Single event handler for DualWield (two SwingFrames).
--
-- We need to predict which one is happening as the combat log does not
-- differentiate between mainhand and offhand swing events.
--
-- Which is most unfortunate, it seems really easy to implement on Blizzard's end.
--
local function SwingFrameDualWieldEvent(self,_,_,event,_,src_guid)
  if ( string.sub(event,1,5) == "SWING" ) and ( src_guid == UnitGUID("player") ) then

    -- this happens when lag throws us out of sync.
    -- there's no easy way to figure out wtf is going on, so just
    -- reset and hope we fall back in sync.
    --
    if ( SwingFrame[self.id].active ) and ( SwingFrame[self.id+1].active ) then
      SwingFrame[self.id].active = nil;
      SwingFrame[self.id+1].active = nil;
    end

    if ( SwingFrame[self.id].active ) then
      SwingFrameReset(SwingFrame[self.id+1]);
      SwingFrame[self.id+1].active = true;
      SwingFrame[self.id+1]:SetScript("OnUpdate", SwingFrameTimer);

    else
      SwingFrameReset(SwingFrame[self.id]);
      SwingFrame[self.id].active = true;
      SwingFrame[self.id]:SetScript("OnUpdate", SwingFrameTimer);
    end
  end
end



-- begin --
-- ----- --

function Whammy.Enable(frame, func)
  SwingFrameReset(frame);
  frame:SetScript("OnEvent", func);
  frame:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED");
  frame:Hide();
end


function Whammy.Disable(frame)
  frame:UnregisterAllEvents();
  frame:Hide();
end


function Whammy.InventoryWatch()

  -- keep for future use if necessary
  --[[
  local i = 0;
  while ( SwingFrame[i] ) do
    Whammy.Disable(SwingFrame[i]);
    i = i + 1;
  end
  ]]--

  Whammy.DualWield = PlayerWeaponIsDualWield();
  Whammy.Ranged = PlayerWeaponIsRanged();

  if ( Whammy.DualWield ) then
    --print("detected dual wield");
    SwingFrame[1].SLOT = Whammy.MAINHAND_SLOT;
    SwingFrame[2].SLOT = Whammy.OFFHAND_SLOT;
    Whammy.Enable(SwingFrame[1], SwingFrameDualWieldEvent);
    Whammy.Enable(SwingFrame[2], nil);
    return;
  else
    SwingFrame[1].SLOT = Whammy.MAINHAND_SLOT;
    Whammy.Disable(SwingFrame[2]);
  end

  if ( Whammy.Ranged ) then
    if ( Whammy.Class == "HUNTER" ) then
      --print("detected ranged (autoshot)");
      Whammy.Enable(SwingFrame[1], SwingFrameRangedEvent);
    elseif ( Whammy.Class == "MAGE" or Whammy.Class == "PRIEST" or Whammy.Class == "WARLOCK" ) then
      --print("detected ranged (wand)");
      Whammy.Enable(SwingFrame[1], SwingFrameRangedEvent);
    end
    -- Rogues and Warriors can equip ranged weapons, but can't use them anymore
  else
    --print("detected melee");
    Whammy.Enable(SwingFrame[1], SwingFrameMainHandEvent);
  end

end


Whammy:SetScript("OnEvent", function(_,event,...)
  --print(event);
  Whammy:UnregisterAllEvents();

  Whammy:SetScript("OnEvent", function(_,event,...)
    --print(event);
    Whammy:InventoryWatch();
  end);

--Whammy:RegisterEvent("PLAYER_TALENT_UPDATE");
  Whammy:RegisterEvent("PLAYER_EQUIPMENT_CHANGED");
  Whammy:InventoryWatch();
end);


-- event debugging
--[[
eventwatch = CreateFrame("FRAME", nil, UIParent);
eventwatch:RegisterAllEvents();
eventwatch:SetScript("OnEvent", function(_,event,...)
  print(event);
end);
]]--

Whammy:RegisterEvent("PLAYER_ENTERING_WORLD");

